/*
 * version.h - version information for the Linux file system degragmenter.
 * version.h,v 1.4 1993/01/07 14:48:51 linux Exp
 *
 * Copyright (C) 1992, 1993 Stephen Tweedie (sct@dcs.ed.ac.uk)
 * 
 * This file may be redistributed under the terms of the GNU General
 * Public License.
 *
 */

char *version = "emergency 0.3b alpha";
char *CVSID = "version.h,v 1.4 1993/01/07 14:48:51 linux Exp";
